/* -*- c-basic-offset: 4 indent-tabs-mode: nil -*-  vi:set ts=8 sts=4 sw=4: */

/*
    Vamp feature extraction plugin using MusicIP OFA library.
    
    See README for more information and licensing terms.
*/

#ifndef _OFA_VAMP_PLUGIN_H_
#define _OFA_VAMP_PLUGIN_H_

#include <vamp-sdk/Plugin.h>
#include <inttypes.h>

class OfaVampPlugin : public Vamp::Plugin
{
public:
    enum Type { TypeFingerprint, TypePUID, TypeBoth };
    OfaVampPlugin(float inputSampleRate, Type type);
    virtual ~OfaVampPlugin();

    bool initialise(size_t channels, size_t stepSize, size_t blockSize);
    void reset();

    InputDomain getInputDomain() const { return TimeDomain; }

    size_t getPreferredStepSize() const;
    size_t getPreferredBlockSize() const;

    size_t getMinChannelCount() const { return 1; }
    size_t getMaxChannelCount() const { return 2; }

    std::string getMaker() const;
    int getPluginVersion() const;
    std::string getCopyright() const;

    ParameterList getParameterDescriptors() const;
    float getParameter(std::string) const;
    void setParameter(std::string, float);

    OutputList getOutputDescriptors() const;

    FeatureSet process(const float *const *inputBuffers,
                       Vamp::RealTime timestamp);

    FeatureSet getRemainingFeatures();

protected:
    Type m_type;
    mutable int m_fingerprintResultIndex;
    mutable int m_puidResultIndex;
    int16_t *m_buffer;
    size_t m_bufsiz;
    size_t m_bufidx;
    size_t m_channels;
    size_t m_blockSize;
    size_t m_totalCount;
    std::string m_print;
    bool m_enough;
};

class OfaFingerprintPlugin : public OfaVampPlugin
{
public:
    OfaFingerprintPlugin(float inputSampleRate) :
        OfaVampPlugin(inputSampleRate, TypeFingerprint) { }

    std::string getName() const { 
        return "MusicIP Audio Fingerprinter";
    }

    std::string getDescription() const { 
        return "Calculates an audio fingerprint using the MusicIP OFA fingerprinting library";
    }

    std::string getIdentifier() const {
        return "ofa_fingerprint";
    }
};

class OfaPUIDPlugin : public OfaVampPlugin
{
public:
    OfaPUIDPlugin(float inputSampleRate) :
        OfaVampPlugin(inputSampleRate, TypePUID) { }

    std::string getName() const { 
        return "MusicIP PUID Lookup";
    }

    std::string getDescription() const { 
        return "Calculates an audio fingerprint using the MusicIP OFA fingerprinting library and uses it to look up a MusicDNS PUID";
    }

    std::string getIdentifier() const {
        return "ofa_puid";
    }
};

class OfaBothPlugin : public OfaVampPlugin
{
public:
    OfaBothPlugin(float inputSampleRate) :
        OfaVampPlugin(inputSampleRate, TypeBoth) { }

    std::string getName() const { 
        return "MusicIP Fingerprinter and PUID";
    }

    std::string getDescription() const { 
        return "Calculates an audio fingerprint and PUID using the MusicIP OFA library";
    }

    std::string getIdentifier() const {
        return "ofa_puid_and_fingerprint";
    }
};


#endif
